{$N+}
{$A-}
unit BigFH ;
{
        Program Name : BigFH
        Package Name : LB
        Purpose      : Big file heap
        Institution  : Conroy & Conroy Co.
        Date Written : 2004
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This is a managed store interface to an array of files.  The
        theoretical maximum heap size is 2,305,843,009,213,693,952 bytes (minus
        allocation table overhead).  However, due to the 2 Gb memory constraint
        under Windows XP, the maximum size could never even reach
        576,460,752,303,423,488 bytes (minus allocation table overhead) or 512
        Pentabytes.
          The store manages one or more file heaps, with each file having up to
        1 Gb of storage.  The maximum theoretical array size is 2 billion
        entries.
}

interface

{$I EDEFINES.INC}

uses { C&C... }
     Collect, { TCollection }
     _DebugIn, { PDebug_Interface }
     MStores, { TManaged_Store }
     _Stores, { TStore_Address }
     _UEHDefs ; { TUEC }

type PSBig_FH = ^TSBig_FH ;
     TSBig_FH = object( TSManaged_Store64 )
                   private { Instance data... }
                       Heaps : TCollection ;
                       Resolution : longint ;
                       Device : byte ;
                       _Filename : string ; { Base filename and path }
                       _Extension : string ; { Filename extension, including dot }
{$IFDEF Debug}
                       Sanity : string[ 8 ] ;
{$ENDIF}

                  public
                      Flags : integer ; { Processing flags.  See FHF_* }
                      Memory : PSManaged_Store ; { Where to allocate cached information }
                      Serial_Number : longint ;
                      Last_Error : TUEC ;

                  public { Constructors and destructors... }
                      constructor Init( Minimum, F : integer ; const Name : string ;
                          var _Result : tUEC ; Mem : PSManaged_Store ) ;
                      destructor Done ; virtual ; { Finish up FH access }

                  private { Internal utility routines }
                      procedure Create_New_Segment( Index : longint ) ;
                      procedure Log_Error( C, S : integer ) ;

                  public { Overrides... }
                      function Extended_Size : TStore_Address64 ; virtual ;

                  public { API... }
                      function Allocate( _Size : TStore_Address64 ) : TStore_Address64 ;
                          virtual ;
                      { Allocate space }

                      function Allocate_At( _Position, _Size : TStore_Address64 ) : boolean ;
                          virtual ;
                      { Allocate space at a specific location }

                      procedure Deallocate( PTR : TStore_Address64 ; _Size : TStore_Address64 ) ;
                          virtual ;
                      { Deallocate space in FH }

                      procedure Read( PTR, Size : TStore_Address64 ;
                          var Buffer ) ; virtual ;
                      procedure Read_Heap( PTR : TStore_Address64 ; _Size : TStore_Address64 ;
                          var Buffer ) ; virtual ; { Read data from FH }
                      function Reallocate( PTR : TStore_Address64 ;
                          Old, _New : TStore_Address64 ) : TStore_Address64 ; virtual ;
                          { Reallocate existing data }

                      procedure Write( PTR : TStore_Address64 ;
                          _Size : TStore_Address64 ; var Buffer ) ; virtual ;
                      procedure Write_Heap( PTR : TStore_Address64 ;
                          _Size : TStore_Address64 ; var Buffer ) ; virtual ;
                      procedure Set_Root( var Buffer ) ; virtual ;
                      { Write data to root block of FH }

                      procedure Get_Root( var Buffer ) ; virtual ;
                      { Read data from root block of FH }

                      function Path : string ; virtual ;

                      function SpaceAvail : TStore_Address64 ; virtual ;
                      function XSpaceAvail : TStore_Address64 ; virtual ;
                      function MaxSpace : TStore_Address64 ; virtual ;
                      function Debugger : PSDebug_Interface ; virtual ;

                      procedure Set_Cache_Flag( Flag : boolean ) ;
                      function Valid_Data_Address( Address : TStore_Address64 ) : boolean ;
                      procedure Write_Cache ;
               end ; { TSBig_FH }

implementation

uses { Borland... }
{$IFDEF DELPHI}
     SysUtils,
{$ENDIF}

     { C&C... }
     Compatib, { FileExists }
     ErrMastr, { Error_Code }
     FSFH, { TFH }
     Num1s, { Num1 }
     O_S, { OS }
     Standard, { Bit_Values }
     TypeDefs ; { Max_File_Size }

const Max_Segment_Size = 1073741824 ;
      Max_Segment_Shift = 30 ;

{$IFDEF Debug}
procedure Check_Big_FH( O : PSBig_FH ) ;

begin
    if( O^.Sanity <> 'TSBig_FH' ) then
    begin
        raise Exception.Create( 'Invalid type-cast' ) ;
        halt ;
    end ;
end ;
{$ENDIF}


{ TSBig_FH methods... }

{ Internal utility routines }

procedure TSBig_FH.Create_New_Segment( Index : longint ) ;

var F : file ;
    FH : PFH ;
    Name : string ;
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    while( Heaps.Count <= Index ) do
    begin
        Heaps.Insert( nil ) ;
    end ;
    if( Heaps.At( Index ) <> nil ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Corrupted_File_Structure, 3 ) ;
        halt ;
    end ;
    Name := _Filename + num1( Index ) + _Extension ;
    if( not FileExists( Name ) ) then
    begin
        Clear_IOResult ; { Cancel any pending errors }
        assign( F, Name ) ;
        {$I-}
        rewrite( F ) ;
        {$I+}
        if( IOResult = 0 ) then
        begin
            {$I-}
            close( F ) ;
            {$I+}
            Clear_IOResult ; { Ignore close errors }
        end ;
    end ;
    FH := System.New( PFH, Init( Resolution, Flags, Name, UEC, Memory ) ) ;
    if( UEC.Code <> 0 ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Corrupted_File_Structure, 3 ) ;
        halt ;
    end ;
    Heaps.AtPut( Index, FH ) ;
end ; { TSBig_FHCreate_New_Segment }


procedure TSBig_FH.Log_Error( C, S : integer ) ;

var
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : tUEC ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Last_Error.Facility := FHErr_Facility ;
    Last_Error.Code := C ;
    R := Error_Code( FHErr_Facility, C ) ;
    UEC := tUEC( R ) ;
    ErrMastr.Log_Error( UEC, '', S ) ;
end ;


{ API... }

constructor TSBig_FH.Init( Minimum, F : integer ; const Name : string ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;
{ Initialize for FH access.  _Result is result of initialization.  F is the
  initialization flags.  Name is the file name.  Minimum is the allocation
  resolution. }

var Dummy : integer ;
    Found : boolean ;

begin
    inherited Init ;

    Heaps.Init( 0, 4 ) ;

{$IFDEF DELPHI}
    System.Initialize( _Filename ) ;
    System.Initialize( _Extension ) ;
{$ENDIF}
{$IFDEF Debug}
    Sanity := 'TSBig_FH' ;
{$ENDIF}

{$IFDEF Verify}
    Address_List := TInteger_List.Create ;
    Size_List := TInteger_List.Create ;
{$ENDIF}
    Memory := Mem ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;

    Flags := F ;
    _Result.Facility := FHErr_Facility ;
    _Result.Code := FHErr_Bad_Minimum ;
    Last_Error := _Result ;
    Found := false ;
    if( Minimum = 0 ) then
    begin
        Minimum := 16 ; { Default resolution }
    end ;
    for Dummy := 0 to 15 do
    begin
        if( Minimum = Bit_Values[ Dummy ] ) then
        begin
            Found := true ;
        end ;
    end ;
    { See if a power of 2 from 1 to 32768. }

    if( not Found ) then
    begin
        exit ;
    end ;
    Resolution := Minimum ;

    Dummy := Extension_Pos( Name ) ;
    if( Dummy = 0 ) then
    begin
        Dummy := length( Name ) + 1 ;
    end ;
    _Extension := System.copy( Name, Dummy, length( Name ) ) ;
    _Filename := System.copy( Name, 1, Dummy - 1 ) ;

    _Result.Facility := FHErr_Facility ;
    _Result.Code := 0 ; { Setup was OK }
    Last_Error := _Result ;
    Serial_Number := Get_Serial_Number ;

    Device := 0 ;
    if( length( Name ) > 1 ) then
    begin
        if( Name[ 2 ] = ':' ) then
        begin
            Device := ord( Name[ 1 ] ) - 64 ;
        end ;
    end ;

    Initialize ;
end ; { TSBig_FHInit }


destructor TSBig_FH.Done ; { Finish up FH access }

var Loop : integer ;

begin
    for Loop := 0 to Heaps.Count - 1 do
    begin
        dispose( PFH( Heaps.At( Loop ) ), Done ) ;
        Heaps.AtPut( Loop, nil ) ;
    end ;
    Heaps.Done ;

{$IFDEF Debug}
    Sanity := ' ' ;
{$ENDIF}
{$IFDEF DELPHI}
    System.Finalize( _Filename ) ;
    System.Finalize( _Extension ) ;
{$ENDIF}

    { Inherited destructor }
    inherited Done ;
end ; { TSBig_FHDone }


function TSBig_FH.Path : string ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Path := _Filename + _Extension ;
end ;


function TSBig_FH.Allocate( _Size : TStore_Address64 ) : TStore_Address64 ;
{ Allocate _Size bytes in store heap and return pointer.  Return 0 if error. }

var FH : PFH ;
    Loop : integer ;
    Res : TStore_Address ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Allocate := 0 ; { Return zero (0) if error }
    if( ( _Size < 1 ) or ( _Size > Max_Segment_Size ) ) then { Cannot allocate nothing or more than maximum }
    begin
        exit ;
    end ;
    for Loop := 0 to Heaps.Count do
    begin
        FH := PFH( Heaps.At( Loop ) ) ;
        if( FH = nil ) then
        begin
            Create_New_Segment( Loop ) ;
            FH := PFH( Heaps.At( Loop ) ) ;
            Allocate := FH^.Allocate( _Size ) ;
            exit ;
        end ;
        Res := FH^.Allocate( _Size ) ;
        if( Res <> 0 ) then { Success }
        begin
            Allocate := Res ;
            exit ;
        end ;
    end ;
    Create_New_Segment( Heaps.Count ) ;
    FH := PFH( Heaps.At( Heaps.Count - 1 ) ) ;
    Allocate := FH^.Allocate( _Size ) ;
end ; { TSBig_FHAllocate }


{ Allocate space at a specific location }
function TSBig_FH.Allocate_At( _Position, _Size : TStore_Address64 ) : boolean ;

var FH : PFH ;
    Index : longint ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Allocate_At := False ; { Return false if error }
    if( ( _Size < 1 ) or ( _Size > Max_Segment_Size ) ) then { Cannot allocate nothing or more than maximum }
    begin
        exit ;
    end ;
    Index := _Position shr Max_Segment_Shift ;
    while( Index >= Heaps.Count ) do { Past end of heap }
    begin
        Heaps.Insert( nil ) ;
    end ;
    FH := Heaps.At( Index ) ;
    if( FH = nil ) then { Unallocated segment }
    begin
        Create_New_Segment( Index ) ;
        FH := Heaps.At( Index ) ;
    end ;
    _Position := _Position and $3FFFFFFF ;
    Allocate_At := FH^.Allocate_At( _Position, _Size ) ;
end ; { TSBig_FHAllocate_At }


procedure TSBig_FH.Deallocate( PTR : TStore_Address64 ; _Size : TStore_Address64 ) ;

var FH : PFH ;
    Index : integer ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    if( ( Flags and FHF_Fast_Deallocate ) <> 0 ) then
    begin
        _Size := 0 ; { Don't deallocate }
    end ;
    Index := PTR shr Max_Segment_Shift ;
    if( Index >= Heaps.Count ) then { Past end of heap }
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        exit ;
    end ;
    FH := Heaps.At( Index ) ;
    if( FH = nil ) then { Unallocated segment }
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        exit ;
    end ;
    PTR := PTR and $3FFFFFFF ;
    FH^.Deallocate( PTR, _Size ) ;
end ; { TSBig_FHDeallocate }


procedure TSBig_FH.Read( PTR, Size : TStore_Address64 ; var Buffer ) ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Read_Heap( PTR, Size, Buffer ) ;
end ;


procedure TSBig_FH.Read_Heap( PTR : TStore_Address64 ; _Size : TStore_Address64 ;
    var Buffer ) ;

var FH : PFH ;
    Index : integer ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    if( _Size = 0 ) then
    begin
        inc( Store_Statistics.Reads ) ;
        exit ; { Do nothing }
    end ;
    Index := PTR shr Max_Segment_Shift ;
    if( Index >= Heaps.Count ) then { Past end of heap }
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        exit ;
    end ;
    FH := Heaps.At( Index ) ;
    if( FH = nil ) then { Unallocated segment }
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        exit ;
    end ;
    PTR := PTR and $3FFFFFFF ;
    FH^.Read_Heap( PTR, _Size, Buffer ) ; { Read data from FH }
end ; { TSBig_FHRead_Heap }


function TSBig_FH.Reallocate( PTR : TStore_Address64 ;
    Old, _New : TStore_Address64 ) : TStore_Address64 ;

var Dummy : longint ;
    FH : PFH ;
    Index : longint ;
    Res, Temp : TStore_Address ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Reallocate := 0 ;
    Dummy := not ( Resolution - 1 ) ;
    if( ( ( Old - 1 ) and Dummy ) = ( ( _New - 1 ) and Dummy ) ) then
    { No actual reallocation is needed: new allocation fits within the old one
      due to minimum resolution. }
    begin
        Reallocate := PTR ;
    end else
    begin
        Index := PTR shr Max_Segment_Shift ;
        if( Index >= Heaps.Count ) then { Past end of heap }
        begin
            Set_Master_Error_Operation( TMEO_Show ) ;
            Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
            exit ;
        end ;
        FH := Heaps.At( Index ) ;
        if( FH = nil ) then { Unallocated segment }
        begin
            Set_Master_Error_Operation( TMEO_Show ) ;
            Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
            exit ;
        end ;
        if( FH^.MaxSpace >= _New ) then { See if we can reallocate within same FH }
        begin
            PTR := PTR and $3FFFFFFF ;
            Res := FH^.Reallocate( PTR, Old, _New ) ;
            if( Res <> 0 ) then
            begin
                Temp := Index ;
                Temp := Temp shl Max_Segment_Shift ;
                Reallocate := Res + Temp ;
            end ;
            exit ;
        end ;
        inherited Reallocate( PTR, Old, _New ) ;
    end ; { if( ( ( Old - 1 ) and Dummy ) = ( ( _New - 1 ) and Dummy ) ) }
end ; { TSBig_FHReallocate }


procedure TSBig_FH.Write( PTR, _Size : TStore_Address64 ;
    var Buffer ) ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Write_Heap( PTR, _Size, Buffer ) ;
end ;


procedure TSBig_FH.Write_Heap( PTR, _Size : TStore_Address64 ; var Buffer ) ;
{ Write data to FH }

var FH : PFH ;
    Index : longint ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    if( _Size = 0 ) then
    begin
        inc( Store_Statistics.Reads ) ;
        exit ; { Do nothing }
    end ;
    Index := PTR shr Max_Segment_Shift ;
    if( Index >= Heaps.Count ) then { Past end of heap }
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        exit ;
    end ;
    FH := Heaps.At( Index ) ;
    if( FH = nil ) then { Unallocated segment }
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        exit ;
    end ;
    PTR := PTR and $3FFFFFFF ;
    FH^.Write_Heap( PTR, _Size, Buffer ) ; { Read data from FH }
end ; { TSBig_FHWrite_Heap }


procedure TSBig_FH.Set_Root( var Buffer ) ; { Write data to root block of FH }

var FH : PFH ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    if( Heaps.Count = 0 ) then
    begin
        Create_New_Segment( 0 ) ;
    end ;
    FH := PFH( Heaps.At( 0 ) ) ;
    FH^.Set_Root( Buffer ) ;
end ;


procedure TSBig_FH.Get_Root( var Buffer ) ; { Read data from root block of FH }

var FH : PFH ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    if( Heaps.Count = 0 ) then
    begin
        Create_New_Segment( 0 ) ;
    end ;
    FH := Heaps.At( 0 ) ;
    if( FH = nil ) then { Heap is empty }
    begin
	fillchar( Buffer, Resolution, 0 ) ;
    end else
    begin
        FH^.Get_Root( Buffer ) ;
    end ;
end ;


function TSBig_FH.XSpaceAvail : TStore_Address64 ;
{ Return available space TStore_Address store heap including file expansion }

var FH : PFH ;
    Loop : longint ;
    R, R1 : TStore_Address ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    R := SpaceAvail ;
    begin
        if( Diskfree( Device ) + R > OS.Max_File_Size * Heaps.Count ) then
        begin
            for Loop := 0 to Heaps.Count - 1 do
            begin
                FH := PFH( Heaps.At( Loop ) ) ;
                R1 := OS.Max_File_Size - FH^.Max_Storage ; { Amount heap can expand }
                R1 := R1 div ( ( Resolution * 8 ) + 1 ) * Resolution * 8 ; { Subtract overhead }
                R := R1 + R ;
            end ;
        end else
        begin
            R := R + Diskfree( Device ) ;
        end ;
    end ;
{$WARNINGS OFF}
    if( R > $7FFFFFFF ) then { Larger value than fits in longint function result }
{$WARNINGS ON}
    begin
        R := $7FFFFFFF ;
    end ;
    XSpaceAvail := R ;
end ; { TSBig_FHXSpaceAvail }


function TSBig_FH.SpaceAvail : TStore_Address64 ;
{ Return available space within store heap without expanding file }

var FH : PFH ;
    Loop : longint ;
    R : TStore_Address ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    R := 0 ;
    for Loop := 0 to Heaps.Count - 1 do
    begin
        FH := Heaps.At( Loop ) ;
        if( FH <> nil ) then
        begin
            R := R + FH^.SpaceAvail ;
        end ;
    end ;
{$WARNINGS OFF}
    if( R > $7FFFFFFF ) then { Larger value than fits in longint function result }
{$WARNINGS ON}
    begin
        R := $7FFFFFFF ;
    end ;
    SpaceAvail := R ;
end ;


function TSBig_FH.MaxSpace : TStore_Address64 ;
{ Return largest contiguous space without extending file. }

var FH : PFH ;
    Loop : longint ;
    R, R1 : longint ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    R := 0 ;
    for Loop := 0 to Heaps.Count - 1 do
    begin
        FH := Heaps.At( Loop ) ;
        if( FH <> nil ) then
        begin
            R1 := FH^.MaxSpace ;
            if( R < R1 ) then
            begin
                R := R1 ;
            end ;
        end ;
    end ;
    MaxSpace := R ;
end ; { TSBig_FHMaxSpace }


function TSBig_FH.Debugger : PSDebug_Interface ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Debugger := nil ; { TODO }
end ;


procedure TSBig_FH.Set_Cache_Flag( Flag : boolean ) ;

var FH : PFH ;
    Loop : longint ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    for Loop := 0 to Heaps.Count - 1 do
    begin
        FH := Heaps.At( Loop ) ;
        if( FH <> nil ) then
        begin
            FH^.Set_Cache_Flag( Flag ) ;
        end ;
    end ;
end ;


function TSBig_FH.Extended_Size : TStore_Address64 ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Extended_Size := XSpaceAvail ;
end ;


function TSBig_FH.Valid_Data_Address( Address : TStore_Address64 ) : boolean ;

var FH : PFH ;
    Index : longint ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    Index := Address shr Max_Segment_Shift ;
    if( Index >= Heaps.Count ) then { Past end of heap }
    begin
        Valid_Data_Address := False ;
        exit ;
    end ;
    FH := Heaps.At( Index ) ;
    if( FH = nil ) then { Unallocated segment }
    begin
        Valid_Data_Address := False ;
        exit ;
    end ;
    Address := Address and $3FFFFFFF ;
    Valid_Data_Address := FH^.Valid_Data_Address( Address ) ;
end ;


procedure TSBig_FH.Write_Cache ;

var FH : PFH ;
    Loop : longint ;

begin
{$IFDEF Debug}
    Check_Big_FH( @self ) ;
{$ENDIF}
    for Loop := 0 to Heaps.Count - 1 do
    begin
        FH := Heaps.At( Loop ) ;
        if( FH <> nil ) then
        begin
            FH^.Write_Cache ;
        end ;
    end ;
end ;



end.
