{
              Program Name : PARSE
              Package Name : LB
              Purpose      : Parsing functions
              Institution  : Conroy & Conroy Co.
              Date Written : 03-Nov-1991
              Written By   : Alan Conroy
              Version      : 2.1A

              Copyright (C) 1991 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *   30-Nov-2002 EAC      Initialize strings for Delphi. *
        *   07-Mar-2006 RBJ      Cleaned up memory leaks.       *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit provides various functions for parsing strings.

}

unit Parse ;

interface

{$I EDEFINES.INC}

uses Common, { TBase_Object }
     Collect, { TInteger_Collection }
{$IFNDEF DELPHI}
     Objects,
{$ENDIF}
     Files, { PSFile }
     UStrings ; { TASCII_String }

function Min_Match( Full, Partial : string ; Len : integer ) : boolean ;
{ Returns true if partial string matches full string, with a minimum of Len
  characters. }

function Parse_Command_Line : string ;
{ This function parses the command line passed to the current program,
  normalizing it based on the DOS switchar parameter, and recognizing both "/"
  and "-" as switch indicators.  Characters within quotes remain unchanged. }

procedure Parse_IO( Command : string ; var Input_Spec, Output_Spec : string ) ;
{ Parse a command line into input and output specifications }

function Parse_Parameter( Separator : string ; var Command : string ) : string ;

function Parse_Quotes( const S : string ) : string ;
{ Parse a quoted string, returning unquoted version }

function Parse_Switch( Switch, Alternate : string ;
    var Command, Parameter : string ) : Integer ;
{ Parse a switch in a command line }

function Switch_Present( Command : string ) : Integer ;
{ Return position of first switch in Command (0 if none) }

function Program_Path : string ;
{ Returns path of executable }

{ Error constants... }
const ParseErr_Success = 0 ;
      ParseErr_Ambiguous_Match = -1 ;

      ParserErr_Success = 0 ;
      ParserErr_Unexpected_End_Of_File = 1 ;
      ParserErr_Unterminated_Literal = 2 ;
      ParserErr_No_Source = 3 ;

type TSParser = object( TBase_Object )
                   private { Instance data... }
		       Command : TASCII_String ; { Current input line }
                       _Current_Line : integer ; { Current line number }

                       Returned_Tokens : Integer ;
		       { Number of returned tokens waiting for reads }

                       Token_Error : Integer ;
		       { Token error returned by Token routine }

		       Token_Stack : TInteger_Collection ;
		       { Stack of token lines }

		       Tokens : TString_Collection ; { Special tokens }

                   public { Constructors and destructors... }
                       constructor Init ;
                       destructor Done ; virtual ;

{$IFDEF DELPHI}
                   protected { Internal utility routines... }
{$ENDIF}
                       function Get_Token : PString ; virtual ;
                       { Return next token from source file }

                       function Pop_Token_Line : Integer ; { Return previous token line}
                       procedure Push_Token_Line ; { Push current token line }
                       function Read_Next_Line : boolean ; virtual ; { Read next line from source data }

                   public { Property handlers... }
                       function Get_Current_Line : longint ;
                       function Get_Token_Line : longint ;

                   public { API... }
                       Non_Terminals : TASCII_String ;
		       Token_Line : Integer ; { Line number of current token }

                       procedure Add_Special_Token( S : PString ) ;
                       function Grab_Line : PString ;
                       function Last_Error : longint ;
                       function Token : PString ;
                       function Token_EOL : boolean ;
                       procedure Put_Token( Value : PString ) ; virtual ;
	       end ;
     PSParser = ^TSParser ;

type TSFile_Parser = object( TSParser )
                        private { Instance data... }
                            _Source : PSFile ; { Input source }

                        public { Constructors and destructors... }
                            constructor Init ;

{$IFDEF DELPHI}
                        protected { Internal utility routines... }
{$ENDIF}
                            function Read_Next_Line : boolean ; virtual ;

                        public { Property handlers... }
                            function Get_Source : PSFile ;
                            procedure Set_Source( Value : PSFile ) ;
                    end ;
     PSFile_Parser = ^TSFile_Parser ;

type TSString_Parser = object( TSParser )
                        private { Instance data... }
                            _Source : TASCII_String ; { Input source }
                            _Supplied : boolean ; { True if the source has already been supplied }

                        public { Constructors and destructors... }
                            constructor Init ;
                            destructor Done ; virtual ;

{$IFDEF DELPHI}
                        protected { Internal utility routines... }
{$ENDIF}
                            function Read_Next_Line : boolean ; virtual ;

                        public { Property handlers... }
                            function Get_Source : PString ;
                            procedure Set_Source( Value : PString ) ;
                    end ;
     PSString_Parser = ^TSString_Parser ;

     TSFiles_Parser = object( TSFile_Parser )
                        public { Constructors and destructors... }
                            constructor Init ;
                            destructor Done ; virtual ;

	                private { Instance data... }
                            _Files : TCollection ;

                        public { API... }
                            procedure Close_File ;
                            procedure Open_File( const Name : string ;
                                F : PSFile ) ;
                    end ;
     PSFiles_Parser = ^TSFiles_Parser ;

{$IFDEF DELPHI}
{ Delphi wrappers for Static parser classes... }

type TParser = class( TBase_Delphi_Object )
                   protected { Instance data... }
                       _Parser : PSParser ;

                   public { Constructors and destructors... }
                       constructor Create ;
                       destructor Destroy ; override ;

                   protected { Internal utility routines... }
                       function Get_Token : string ; virtual ;
                       { Return next token from source file }

                       function Previous_Token_Line : Integer ; { Return previous token line}
                       procedure Push_Token_Line ; { Push current token line }
                       function Read_Next_Line : boolean ; virtual ; { Read next line from source data }

                   public { Property handlers... }
                       function Get_Current_Line : longint ;
                       function Get_Token_Line : longint ;
                       function Get_Non_Terminals : string ;
                       procedure Set_Non_Terminals( S : string ) ;
                       function _Get_Token_Line : integer ;
                       procedure Set_Token_Line( V : integer ) ;

                   public { API... }
                       property Non_Terminals : string
                           read Get_Non_Terminals
                           write Set_Non_Terminals ;
		       property Token_Line : Integer
                           read _Get_Token_Line
                           write Set_Token_Line ;

                       procedure Add_Special_Token( S : string ) ;
                       function Grab_Line : string ; virtual ;
                       function Last_Error : longint ;
                       procedure Put_Token( const S : string ) ; virtual ;
                       function Token : string ; virtual ;
                       function Token_EOL : boolean ; virtual ;
	       end ;

     TFile_Parser = class( TParser )
                        public { Constructors and destructors... }
                            constructor Create ;

                        public { Property handlers... }
                            function Get_Source : PSFile ;
                            procedure Set_Source( Value : PSFile ) ;
                    end ;

     TString_Parser = class( TParser )
                        public { Constructors and destructors... }
                            constructor Create ;

                        public { Property handlers... }
                            function Get_Source : string ;
                            procedure Set_Source( Value : string ) ;
                    end ;

     TFiles_Parser = class( TFile_Parser )
                        public { Constructors and destructors... }
                            constructor Create ;

                        public { API... }
                            procedure Close_File ;
                            procedure Open_File( Name : string ; F : PSFile ) ;
                    end ;
{$ENDIF}

var Parse_Error : integer ;

implementation

uses ASCIIDef, { CR }
     CommonUt, { Edit }
     Compatib, { uppercase }
     O_S, { OS }
     _UEHDefs, { TUEC }
     Standard,
     Instrs ;

{ Unit API... }

function Min_Match( Full, Partial : string ; Len : integer ) : boolean ;

begin
    Parse_Error := ParseErr_Success ;
    if( copy( Full, 1, length( Partial ) ) = Partial ) then
    begin
	if( length( Partial ) >= Len ) then
	begin
	    Min_Match := True ;
	end else
	begin
	    Min_Match := False ;
	    Parse_Error := ParseErr_Ambiguous_Match ;
	end ;
    end else
    begin
	Min_Match := False ;
    end ;
end ;


function Parse_Command_Line : string ; { Return parsed command line }

var  Command, Command1 : string ; { Command line }
     Dummy1 : Integer ; { Loop variable }
     Quote : Char ; { Current quote }

begin
    Parse_Error := ParseErr_Success ;
    Command := Edit( OS^.Command_Line, 8 ) ; { Get command line }
    Command1 := OS^.Switch_Delimiters( 0 ) ;
    if( ( Command1 <> '"' ) and ( Command1 <> #39 ) and ( Command1 <> '/' ) and
	( Command1 <> '-' ) ) then
    begin
	Quote := ' ' ;
	Dummy1 := 1 ;
	while Dummy1 <= Length( Command ) do { Process entire string }
	begin
	    if( Quote <> ' ' ) { if within a quote }
		then
		    if( Command[ Dummy1 ] = Quote ) then Quote := ' ' else
		else { Not within a quote }
		    if ( Command[ Dummy1 ] = '"' ) or ( Command[ Dummy1 ] = Chr( 39 ) )
			then Quote := Command[ Dummy1 ] ; { Check for quotes }
	    if( Quote = ' ' ) then
            begin
                if( Command[ Dummy1 ] = Command1 ) then
                begin
                    Command[ Dummy1 ] := '/' ;
                end ;
            end ;
	    Dummy1 := Dummy1 + 1 ; { Move to next character }
	end ;
    end ;
    Parse_Command_Line := Command ; { Return line }
end ;


procedure Parse_IO( Command : string ; var Input_Spec, Output_Spec : string ) ;
{ Parse a command line into input and output specifications }

var Dummy : integer ;
    Space_Count, Equal_Count : integer ;
    Space_Pos, Equal_Pos : integer ;
    Quote_Type : char ;

begin
    Parse_Error := ParseErr_Success ;
    Space_Count := 0 ;
    Equal_Count := 0 ;
    Equal_Pos := 0 ;
    Space_Pos := 0 ;
    Quote_Type := ' ' ;
    for Dummy := 1 to length( Command ) do
    begin
	if( Quote_Type <> ' ' ) then { Inside a quote }
	begin
	    if( Command[ Dummy ] = Quote_Type ) then
	    begin
		Quote_Type := ' ' ; { Done with quote }
	    end ;
	end else
	begin
	    if( ( Command[ Dummy ] = '"' ) or ( Command[ Dummy ] = #39 ) ) then
	    begin
		Quote_Type := Command[ Dummy ] ;
	    end else
	    begin
		if( Command[ Dummy ] = ' ' ) then
		begin
		    inc( Space_Count ) ;
		    Space_Pos := Dummy ;
		end ;
		if( Command[ Dummy ] = '=' ) then
		begin
		    inc( Equal_Count ) ;
		    Equal_Pos := Dummy ;
		end ;
	    end ;
	end ;
    end ; { for Dummy := 1 to length( Command ) }
    if( ( Equal_Count = 0 ) and ( Space_Count = 0 ) ) then { Only input present }
    begin
	Input_Spec := Command ;
	Output_Spec := '' ;
	exit ;
    end ;
    if(
	(
	  ( Equal_Count > 0 )
	  and
	  ( Space_Count > 0 )
	)
	or
	( Equal_Count > 1 )
	or
	( Space_Count > 1 )
      ) then
    begin
	Input_Spec := '' ;
	Output_Spec := '' ; { Invalid specification }
	exit ;
    end ;
    if( Space_Count = 1 ) then
    begin
	Input_Spec := copy( Command, 1, Space_Pos - 1 ) ;
	Output_Spec := copy( Command, Space_Pos + 1, 254 ) ;
    end else
    begin
	Output_Spec := copy( Command, 1, Equal_Pos - 1 ) ;
	Input_Spec := copy( Command, Equal_Pos + 1, 254 ) ;
    end ;
end ; { Parse_IO }


function Parse_Parameter( Separator : string ; var Command : string ) : string ;
{ Return next parameter }

var Dummy : Integer ;
    Quote : Char ;

begin
    Parse_Error := ParseErr_Success ;
    Quote := ' ' ; { No quote yet }
    if Length( Separator ) = 0 then Separator := ' ' ; { Default to space }
    for Dummy := 1 to Length( Command ) do
    begin
	if( Quote <> ' ' ) { if within a quote }
	    then
		if( Command[ Dummy ] = Quote ) then Quote := ' ' else
	    else { Not within a quote }
		if( ( Command[ Dummy ] = '"' ) or ( Command[ Dummy ] = Chr( 39 ) ) )
		    then Quote := Command[ Dummy ] ; { Check for quotes }
	if( Command[ Dummy ] = Separator[ 1 ] ) then if( Quote = ' ' ) then
	begin
	    Parse_Parameter := Copy( Command, 1, Dummy - 1 ) ; { Return parameter }
	    Command := Copy( Command, Dummy + 1, 255 ) ; { Strip from command line }
	    Exit ; { Done }
	end ;
    end ;
    { At this point, the entire command line is a parameter }
    Parse_Parameter := Command ; { Return next parameter }
    Command := '' ; { Strip from command line }
end ; { Parse_Parameter }


function Parse_Quotes( const S : string ) : string ;
{ Parse a quoted string, returning unquoted version }

var Dummy : integer ;
    Ch, Quote : char ;
    Res : string ;

begin
    Res := '' ;
    Quote := ' ' ;
    for Dummy := 1 to length( S ) do
    begin
        Ch := S[ Dummy ] ;
        if( Quote = ' ' ) then { Not within a quote }
        begin
            if( ( Ch = #39 ) or ( Ch = '"' ) ) then
            begin
                Quote := Ch ; { Starting a quote }
            end else
            begin
                Res := Res + Ch ;
            end ;
        end else
        begin
            if( Ch = Quote ) then { End of current quote }
            begin
                Quote := ' ' ;
            end else
            begin
                Res := Res + Ch ;
            end ;
        end ;
    end ;
    Parse_Quotes := Res ;
end ; { Parse_Quotes }


function Parse_Switch( Switch, Alternate : string ;
    var Command, Parameter : string ) : Integer ;
{ Parse a switch in a command line:
  Switch is a list of switches, separated by commas,
  Alternate is a list of switches, separated by commas,
  Command is the command line to search for the switches in,
  Parameter is the parameter following the switch (if any).

  Function returns:
	0 = No matching switches found
	1 = Switch matching one in Switch was found
	2 = Switch matching one in Alternate was found

  In the case of multiple switches, only the last occurance is parsed, but all
  occurances are removed from the command line.  Note that switches are passed
  without slashes.  Switch lists may optionally contain vertical bars (|) to
  indicate the minimum abbreviation allowed for the switch. }

label Try_Again ;

var Switches : array[ 0..15 ] of string ;
    Switch_Min : array[ 0..15 ] of Integer ;
    Alternates : array[ 0..15 ] of string ;
    Alt_Min : array[ 0..15 ] of Integer ;
    Dummy, Dummy1, Dummy2 : Integer ; {Work variables}
    First_Time : Boolean ; {True if first switch match}
    Location : Integer ; {Location of switch in Work}
    Quote : Char ;
    Work, Work1 : string ;

begin
    Parse_Error := ParseErr_Success ;

    { The following code breaks the multiple switches in Switch and Alternate
      into two arrays of switches to make the following processing easier }
    for Dummy := 0 to 15 do
    begin
	Switches[ Dummy ] := '' ;
	Alternates[ Dummy ] := '' ;
    end ;
    Dummy1 := -1 ;
    while( Length( Switch ) > 0 ) do
    begin
	Dummy := Pos( ',', Switch + ',' ) ; { Find end of next switch }
	Dummy1 := Dummy1 + 1 ;
	if( Dummy1 = 16 ) then
        begin
	    Switch := '' ; { 16 is the maximum }
        end else
        begin
            Switches[ Dummy1 ] := Copy( Switch, 1, Dummy - 1 ) ; { Get switch }
            Switch := Copy( Switch, Dummy + 1, 255 ) ; { Strip from parameter }
            Dummy2 := Pos( '|', Switches[ Dummy1 ] ) ; { Abbreviation specification? }
            if( Dummy2 > 1 ) then { Abbreviation specified }
            begin
                Switch_Min[ Dummy1 ] := Dummy2 - 1 ;
                Switches[ Dummy1 ] := Copy( Switches[ Dummy1 ], 1, Dummy2 - 1 ) +
                    Copy( Switches[ Dummy1 ], Dummy2 + 1, 255 ) ;
            end else
            begin
                Switch_Min[ Dummy1 ] := Length( Switches[ Dummy1 ] ) ;
            end ;
        end ;
    end ;
    Dummy1 := -1 ;
    while( Length( Alternate ) > 0 ) do
    begin
	Dummy := Pos( ',', Alternate + ',' ) ; { Find end of next switch }
	Dummy1 := Dummy1 + 1 ;
	if( Dummy1 = 16 ) then
        begin
            Alternate := '' ; { 16 is the maximum }
        end else
        begin
            Alternates[ Dummy1 ] := Copy( Alternate, 1, Dummy - 1 ) ; { Get switch }
            Alternate := Copy( Alternate, Dummy + 1, 255 ) ; { Strip from parameter }
            Dummy2 := Pos( '|', Alternates[ Dummy1 ] ) ;
            { Abbreviation specification? }
            if( Dummy2 > 1 ) then { Abbreviation specified }
            begin
                Alt_Min[ Dummy1 ] := Dummy2 - 1 ;
                Alternates[ Dummy1 ] := Copy( Alternates[ Dummy1 ], 1, Dummy2 - 1 ) +
                    Copy( Alternates[ Dummy1 ], Dummy2 + 1, 255 ) ;
            end else
            begin
                Alt_Min[ Dummy1 ] := Length( Alternates[ Dummy1 ] ) ;
            end ;
        end ;
    end ;

    Work := Command ; { Make a local copy of the command line }
    Parameter := '' ; { Thus, no parameter }
    Parse_Switch := 0 ; { No match }
    First_Time := True ; { First time }
Try_Again:
    if( Length( Work ) = 0 ) then
    begin
        Exit ; { No more switches in command line }
    end ;
    Dummy := Pos( '"', Work + '"' ) ; { Find location of quote }
    Dummy1 := Pos( Chr( 39 ), Work + Chr( 39 ) ) ; { Try alternate quote }
    if( Dummy1 < Dummy ) then
    begin
        Dummy := Dummy1 ; { The first quote in the line }
    end ;
    if( Dummy = 1 ) then { Command line starts with a literal }
    begin
	Dummy1 := Instr( Dummy + 1, Work + Work[ 1 ], Work[ 1 ] ) ;
	{ Find ending quote }
	Work := Copy( Work, Dummy1 + 1, 255 ) ; { Strip literal from work string }
	goto Try_Again ; { Go back and try it again }
    end ;

    { At this point, Work contains the rightmost n characters in Command and
      Dummy is the location of the first quote in Work. }

    Dummy1 := Length( Work ) + 1 ;
    for Dummy2 := 0 to 15 do { Try all possible switches }
    begin
	if Length( Switches[ Dummy2 ] ) > 0 then
	begin
	    Location := Pos( '/' + Copy( Switches[ Dummy2 ], 1, Switch_Min[ Dummy2 ] ),
		uppercase( Work ) + '/' + Switches[ Dummy2 ] ) ; { Find one }
	    if( Dummy1 > Location ) then
	    begin
		Dummy1 := Location ; { Choose earliest one }
		Work1 := Switches[ Dummy2 ] ;
		Parse_Switch := 1 ;
	    end ;
	end ;
	if( Length( Alternates[ Dummy2 ] ) > 0 ) then
	begin
	    Location := Pos( '/' + Copy( Alternates[ Dummy2 ], 1, Alt_Min[ Dummy2 ] ),
		uppercase( Work ) + '/' + Alternates[ Dummy2 ] ) ; { Find one }
	    if( Dummy1 > Location ) then { Choose earliest one }
	    begin
		Dummy1 := Location ;
		Work1 := Alternates[ Dummy2 ] ;
		Parse_Switch := 2 ;
	    end ;
	end ;
    end ;

    { At this point, Dummy1 contains the location of the first occurance }

    if( Dummy1 > Length( Work ) ) then { No matching switches found }
    begin
	if( First_Time ) then
        begin
            Parse_Switch := 0 ; { No match }
        end ;
	Exit ;
    end ;
    if( Dummy1 > Dummy ) then { Switch begins after a quote }
    begin
	Quote := Work[ Dummy ] ;
	Dummy := Pos( Quote, Work + Quote ) ; { Find end of literal }
	Work := Copy( Work, Dummy + 1, 255 ) ; { Strip literal }
	if( First_Time ) then
        begin
            Parse_Switch := 0 ; { No match }
        end ;
	goto Try_Again ;
    end ;
    { Found a possible switch location at Dummy1 }
    Dummy1 := Dummy1 + 1 ; { Skip slash }
    Dummy2 := 0 ; { Offset from switch beginning }
    while(
           ( Dummy1 + Dummy2 <= Length( Work ) )
           and
           ( Work[ Dummy1 + Dummy2 ] <> ':' )
           and
           ( Work[ Dummy1 + Dummy2 ] <> '=' )
           and
	   ( Work[ Dummy1 + Dummy2 ] <> '/' )
           and
	   ( Work[ Dummy1 + Dummy2 ] <> ',' )
	   and
	   ( Work[ Dummy1 + Dummy2 ] <> ' ' )
           and
           ( Work[ Dummy1 + Dummy2 ] <> '	' )
         ) do
    begin
        Dummy2 := Dummy2 + 1 ; { Move to next character }
    end ;
    { At this point, Dummy1 is the start of the switch and Dummy2 is the size }
    if( Copy( uppercase( Work ), Dummy1, Dummy2 ) <> Copy( Work1, 1, Dummy2 ) ) then
    { Incomplete match }
    begin
	Work := Copy( Work, Dummy1 + Dummy2, 255 ) ; { Strip switch }
	if( First_Time ) then
        begin
            Parse_Switch := 0 ; { Not a match }
        end ;
	goto Try_Again ; { Try again }
    end ;
    First_Time := False ;

    { At this point, we have a match - now grab any parameters }
    Command := Copy( Command, 1, Length( Command ) - Length( Work ) + Dummy1 - 2 ) +
	Copy( Work, Dummy1 + Dummy2, 255 ) ; { Strip switch from command line }
    Work := Copy( Work, Dummy1 + Dummy2, 255 ) ; { and from work variable }
    if(
        ( Length( Work ) = 0 )
        or
        (
          ( Work[ 1 ] <> ':' )
          and
          ( Work[ 1 ] <> '=' )
        )
      ) then { No parameters }
    begin
	Parameter := '' ;
	goto Try_Again ;
    end ;
    Command := Copy( Command, 1, Length( Command ) - Length( Work ) ) + Copy( Work, 2, 255 ) ;
    { Remove separator }
    Work := Copy( Work, 2, 255 ) ; { Ditto }
    if( ( Work[ 1 ] = '"' ) or ( Work[ 1 ] = Chr( 39 ) ) ) then { Quoted literal }
    begin
	Dummy := Instr( 2, Work + Work[ 1 ], Work[ 1 ] ) ; { Find end of quote }
    end else
    begin
        Dummy := Pos( '/', Work + '/' ) - 1 ; { Find end of parameter }
        if( Pos( ' ', Work + ' ' ) - 1 < Dummy ) then
        begin
            Dummy := Pos( ' ', Work + ' ' ) - 1 ; { Find end of parameter }
        end ;
        if( Pos( ' ', Work + ' ' ) - 1 < Dummy ) then
        begin
            Dummy := Pos( '	', Work + '	' ) - 1 ; { Find end of parameter }
        end ;
    end ;
    Parameter := Copy( Work, 1, Dummy ) ; { Get the parameter }
    if(
        ( copy( Parameter, 1, 1 ) = '"' )
        or
        ( copy( Parameter, 1, 1 ) = #39 )
      ) then { Quoted }
    begin
        if( copy( Parameter, 1, 1 ) = copy( Parameter, length( Parameter ), 1  ) ) then
        begin
            Parameter := copy( Parameter, 1, length( Parameter ) - 1 ) ; { Remove terminating quote }
        end ;
        Parameter := copy( Parameter, 2, length( Parameter ) ) ; { Remove starting quote }
    end ;
    Command := Copy( Command, 1, Length( Command ) - Length( Work ) ) +
        Copy( Work, Dummy + 1, 255 ) ; { Strip parameter from command line }
    Work := Copy( Work, Dummy + 1, 255 ) ; { Remove parameter }
    goto Try_Again ; { Now see if a later switch over-rides this occurance }
end ; { Parse_Switch }


function Switch_Present( Command : string ) : Integer ;
{ Return position of first switch in Command (0 if none) }

var Dummy : Integer ;
    Quote : Char ;

begin
    Parse_Error := ParseErr_Success ;
    Switch_Present := 0 ;
    Quote := ' ' ;
    Dummy := 1 ;
    while( Dummy <= Length( Command ) ) do {Process entire string}
    begin
	if( Quote <> ' ' ) then {if within a quote}
        begin
            if( Command[ Dummy ] = Quote ) then
            begin
                Quote := ' ' ;
            end ;
        end else { Not within a quote }
        begin
            if( ( Command[ Dummy ] = '"' ) or ( Command[ Dummy ] = Chr( 39 ) ) ) then
            begin
                Quote := Command[ Dummy ] ; { Check for quotes }
            end ;
        end ;
	if( Quote = ' ' ) then if( Command[ Dummy ] = '/' ) then
	begin
	    Switch_Present := Dummy ; { Switch found }
	    Exit ;
	end ;
	Dummy := Dummy + 1 ; { Move to next character }
    end ;
end ; { Switch_Present }


function Program_Path : string ;

var Dummy : integer ;
    S : string ;

begin
    S := paramstr( 0 ) ;
    Dummy := length( S ) ;
    while( ( Dummy > 0 ) and ( S[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    Program_Path := copy( S, 1, Dummy )  ;
end ;


{ TSFile_Parser methods... }

{ Constructors and destructors... }

constructor TSFile_Parser.Init ;

begin
    inherited Init ;

    _Source := nil ;
end ;


{ Internal utility routines... }

function TSFile_Parser.Read_Next_Line : boolean ;

var EF : boolean ;
    Res : TUEC ;
    S : string ;

begin
    if( ( _Source = nil ) and ( Command.Length = 0 ) ) then
    begin
        Token_Error := ParserErr_No_Source ;
        Read_Next_Line := False ;
        exit ;
    end ;

    Ef := False ;
{
    if( Get_Source^.EOF ) then
    begin
        EF := True ;
    end ;
}
    {$I-}
    Get_Source^.Readln( S ) ;
    {$I+}
    Command.Set_String( S ) ;
    Res := Get_Source^.IO_Error ;
    if( ( Res.Code <> 0 ) or EF ) then { EOF }
    begin
        Command.Assign( nil ) ;
        Read_Next_Line := False ;
        Token_Error := ParserErr_Unexpected_End_Of_File ; { Unexpected end of file }
        Exit ;
    end ;
    Read_Next_Line := True ;
end ; { TSFile_Parser.Read_Next_Line }


function TSFile_Parser.Get_Source : PSFile ;

begin
    Get_Source := _Source ;
end ;


procedure TSFile_Parser.Set_Source( Value : PSFile ) ;

begin
    _Source := Value ;
end ;


{ TSParser methods... }

{ Constructors and destructors... }

constructor TSParser.Init ;

begin
    inherited Init ;

    Tokens.Init( 0, 1 ) ;
    Token_Stack.Init( 0, 4 ) ;
    Command.Init ;
    _Current_Line := 0 ;
    Token_Error := 0 ;
    Returned_Tokens := 0 ;
    Token_Line := 0 ;
    Non_Terminals.Init ;
    Non_Terminals.Set_String( '$_ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' ) ;
end ;


destructor TSParser.Done ;

begin
    Tokens.Done ;
    Token_Stack.Done ;
    Command.Done ;

    Non_Terminals.Done ;

    inherited Done ;
end ;


{ Internal utility routines... }

function TSParser.Get_Token : PString ; { Return next token from source }

var Aa : integer ; { Changed from byte to support lines > 255 bytes in length }
    Bb : Char ;
    C : PASCII_String ;
    Quote : Boolean ;
    Quote_Type : Char ;
    Temp : PString ;

Label Next, Token_End ;

begin
    Token_Error := ParserErr_Success ;
    Quote_Type := ' ' ;

Next:
    if( Command.Length = 0 ) then { No more text in buffer }
    begin
        if( not Read_Next_Line ) then
        begin
            Get_Token := new( PASCII_String, Init ) ;
            exit ;
        end ;
        Inc( _Current_Line ) ; { Increment line count }
	    Token_Line := Get_Current_Line ; { Token line is now same as current line }

        Temp := Command.Edit( 8700 ) ; { Format line }
        Command.Assign( Temp ) ;
        dispose( Temp, Done ) ;
        if( Command.Length = 0 ) then
        begin
            goto Next ; { Ignore blank lines }
        end ;
    end ;
    Quote := False ;
    C := new( PASCII_String, Init ) ;
    for Aa := 1 to Command.Length do
    begin
        Bb := char( Command.Upcase( Command.At( Aa ) ) ) ; { Get current character }
        if Quote then { Within a quoted token }
        begin
            if( Bb = Quote_Type ) then { End of quoted token }
            begin
                Temp := Command.Copy( 1, Aa ) ;
                C^.Assign( Temp ) ;
                dispose( Temp, Done ) ;
                Temp := Command.Copy( Aa + 1, Command.Length - Aa ) ;
                Command.Assign( Temp ) ;
                dispose( Temp, Done ) ;
                goto Token_End ;
            end ;
        end else
        if( Aa = 1 ) then { First character of token }
        begin
            if( ( Bb = '"' ) or ( Bb = Chr( 39 ) ) ) then { if a quote }
            begin
                Quote := True ; { Now within quotes }
                Quote_Type := Bb { Save quote type }
            end else
            begin
                if( Non_Terminals.Instr( 1, Temp_String( Bb ) ) = 0 ) then
                begin
                    C^.Set_String( char( Command.At( 1 ) ) ) ;
                    Temp := Command.Copy( 2, Command.Length - 1 ) ;
                    Command.Assign( Temp ) ;
                    dispose( Temp, Done ) ;
                    goto Token_End ;
                end ;
            end ;
        end else { After first character }
        begin
            if( Non_Terminals.Instr( 1, Temp_String( Bb ) ) = 0 ) then
            begin
                Temp := Command.Copy( 1, Aa - 1 ) ;
                C^.Assign( Temp ) ;
                dispose( Temp, Done ) ;
                Temp := Command.Copy( Aa, Command.Length - Aa + 1 ) ;
                Command.Assign( Temp ) ;
                dispose( Temp, Done ) ;
                goto Token_End ;
            end ;
        end ;
    end ; { for Aa := 1 to Command.Length }
    C^.Assign( @Command ) ; { Token runs to end of line }
    Command.Assign( nil ) ;
    if( Quote ) then
    begin
        Token_Error := ParserErr_Unterminated_Literal ;
    end ;
Token_End:
    Temp := C^.Edit( -1 ) ; { Format token }
    C^.Assign( Temp ) ;
    dispose( Temp, Done ) ;
    if( C^.Length = 0 ) then
    begin
        // need to dispose of PASCII_String
        dispose( C, Done ) ;

        goto Next ; { Ignore spaces, etc }
    end ;
{    Push_Token_Line ;}
    if( Returned_Tokens > 0 ) then
    begin
        Token_Line := Pop_Token_Line ;
	    Returned_Tokens := Returned_Tokens - 1 ;
    end ;
    Get_Token := C ; { Return value }
end ; { TSParser.Get_Token }


function TSParser.Pop_Token_Line : Integer ; { Return previous token line }

begin
    if( Token_Stack.Count = 0 ) then
    begin
        Pop_Token_Line := 0 ;
    end else
    begin
        Pop_Token_Line := Token_Stack.At( Token_Stack.Count - 1 ) ;
        Token_Stack.AtDelete( Token_Stack.Count - 1 ) ;
    end ;
end ;


procedure TSParser.Push_Token_Line ; { Push current token line }

begin
    Token_Stack.Insert( Token_Line ) ;
end ;


function TSParser.Read_Next_Line : boolean ;

begin
    Command.Assign( nil ) ;
    Read_Next_Line := False ;
end ;


{ Property handlers... }

function TSParser.Get_Token_Line : longint ;

begin
    Get_Token_Line := Token_Line ;
end ;


function TSParser.Get_Current_Line : longint ;

begin
    Get_Current_Line := _Current_Line ;
end ;


{ API... }

procedure TSParser.Add_Special_Token( S : PString ) ;

begin
    Tokens.Insert( S^.Text ) ;
end ;


function TSParser.Grab_Line : PString ;

var Temp : PString ;

label Next ;

begin
    if( Token_Stack.Count > 0 ) then
    begin
        Token_Line := Token_Stack.At( Token_Stack.Count - 1 ) ;
    end ;
    Returned_Tokens := 0 ;
    while( Token_Stack.Count > 0 ) do
    begin
        Token_Stack.AtDelete( Token_Stack.Count - 1 ) ;
    end ;
    if( Command.Length = 0 ) then { No more text in buffer }
    begin
Next:
        if( Read_Next_Line ) then
        begin
            Inc( _Current_Line ) ; { Increment line count }
            Token_Line := Get_Current_Line ; { Token line is now same as current line }

            Temp := Command.Edit( 8700 ) ; { Format line }
            Command.Assign( Temp ) ;
            dispose( Temp, Done ) ;
            if( Command.Length = 0 ) then
            begin
                goto Next ; { Ignore blank lines }
            end ;
        end ;
    end ;
    Temp := new( PASCII_String, Init ) ;
    Temp^.Assign( @Command ) ;
    Grab_Line := Temp ;
    Command.Assign( nil ) ;
end ; { TSParser.Grab_Line }


function TSParser.Last_Error : longint ;

begin
    Last_Error := Token_Error ;
end ;


function TSParser.Token : PString ;

var Dummy : integer ;
    S, T : PString ; { Accumulated token }
    Temp : PString ;

begin
    T := Get_Token ;
    if( ( Tokens.Count > 0 ) and ( T^.Length > 0 ) ) then
    begin
        { A list of special tokens is present and we have a token }
	    for Dummy := 0 to Tokens.Count - 1 do
	    begin
	        S := Temp_String( Tokens.At( Dummy ) ) ;
	        if( T^.At( 1 ) = S^.At( 1 ) ) then { A potential match }
	        begin
                Temp := Command.Copy( 1, S^.Length - 1 ) ;
		        if( T^.Text + Temp^.Text = S^.Text ) then { Found a match }
		        begin
                    dispose( Temp, Done ) ;
		            while( T^.Length < S^.Length ) do
		            begin
			            T^.Append( Get_Token ) ;
		            end ;
		            Token := T ;
		            exit ;
		        end ;
                dispose( Temp, Done )
	        end ;
	    end ;
    end ;
    Token := T ; { Return result }
end ; { TSParser.Token }


procedure TSParser.Put_Token( Value : PString ) ;

var Temp : PASCII_String ;

begin
    if( Value^.Length > 0 ) then
    begin
        Inc( Returned_Tokens ) ;
        Temp := new( PASCII_String, Init ) ;
        Temp^.Assign( @Command ) ;
        if( ( Command.Length > 0 ) and ( Command.At( 1 ) <> 32 ) ) then
        begin
            Command.Assign( Value ) ;
            Command.Append_Char( 32 ) ;
            Command.Append( Temp ) ;
        end else
        begin
            Command.Assign( Value ) ;
            Command.Append( Temp ) ;
        end ;
        dispose( Temp, Done ) ;
        Push_Token_Line ;
{        Token_Line := Token_Stack.At( Token_Stack.Count - 1 ) ;}
    end ;
end ;


function TSParser.Token_EOL : boolean ;

begin
    Token_EOL := ( Command.Length = 0 ) ;
end ;


{ TSString_Parser methods... }

{ Constructors and destructors... }

constructor TSString_Parser.Init ;

begin
    inherited Init ;

    _Source.Init ;
    _Supplied := False ;
end ;


destructor TSString_Parser.Done ;

begin
    _Source.Done ;

    inherited Done ;
end ;



{ Internal utility routines... }

function TSString_Parser.Read_Next_Line : boolean ;

begin
    if( _Supplied ) then
    begin
        Command.Assign( nil ) ;
        Read_Next_Line := False ;
    end else
    begin
        _Supplied := True ;
        Command.Assign( @_Source ) ;
        Read_Next_Line := True ;
    end ;
end ;


{ Property handlers... }

function TSString_Parser.Get_Source : PString ;

var Temp : PASCII_String ;

begin
    Temp := new( PASCII_String, Init ) ;
    Temp^.Assign( @_Source ) ;
    Get_Source := Temp ;
end ;


procedure TSString_Parser.Set_Source( Value : PString ) ;

begin
    _Source.Assign( Value ) ;
    Command.Assign( @_Source ) ;
    _Supplied := True ;
end ;


{ TSFiles_Parser methods... }

{ Constructors and destructors... }

constructor TSFiles_Parser.Init ;

begin
    inherited Init ;

    _Files.Init( 0, 4 ) ;
end ;


destructor TSFiles_Parser.Done ;

begin
    _Files.Done ;

    inherited Done ;
end ;


{ API... }

procedure TSFiles_Parser.Close_File ;

var F : PSFile ;

begin
    F := _Files.At( _Files.Count - 1 ) ;
    dispose( F, Done ) ;
    _Files.AtDelete( _Files.Count - 1 ) ;
    if( _Files.Count = 0 ) then
    begin
        F := nil ;
    end else
    begin
        F := _Files.At( _Files.Count - 1 ) ;
    end ;
    Set_Source( F ) ;
end ;


procedure TSFiles_Parser.Open_File( const Name : string ; F : PSFile ) ;

var UEC : TUEC ;

begin
    if( F = nil ) then
    begin
        F := Files.Open( Name, FSOM_Input, UEC ) ;
    end ;
    _Files.Insert( F ) ;
    Set_Source( F ) ;
end ;


{$IFDEF DELPHI}
{ TParser methods... }

{ Constructors and destructors... }
constructor TParser.Create ;

begin
    inherited Create ;

    _Parser := new( PSParser, Init ) ;
end ;


destructor TParser.Destroy ;

begin
    dispose( _Parser, Done ) ;

    inherited Destroy ;
end ;


{ Internal utility routines... }

function TParser.Get_Token : string ;

var Temp : PString ;

begin
    Temp := _Parser^.Get_Token ;
    Result := Temp^.Text ;
    dispose( Temp, Done ) ;
end ;


function TParser.Previous_Token_Line : Integer ;

begin
    Result := _Parser^.Pop_Token_Line ;
end ;


procedure TParser.Push_Token_Line ;

begin
    _Parser^.Push_Token_Line ;
end ;


function TParser.Read_Next_Line : boolean ;

begin
    Result := _Parser^.Read_Next_Line ;
end ;


{ Property handlers... }

function TParser.Get_Current_Line : longint ;

begin
    Result := _Parser^.Get_Current_Line ;
end ;


function TParser.Get_Token_Line : longint ;

begin
    Result := _Parser^.Get_Token_Line ;
end ;


function TParser.Get_Non_Terminals : string ;

begin
    Result := _Parser^.Non_Terminals.Text ;
end ;


procedure TParser.Set_Non_Terminals( S : string ) ;

begin
    _Parser^.Non_Terminals.Set_String( S ) ;
end ;


function TParser._Get_Token_Line : integer ;

begin
    Result := _Parser^.Token_Line ;
end ;


procedure TParser.Set_Token_Line( V : integer ) ;

begin
    _Parser^.Token_Line := V ;
end ;


{ API... }

procedure TParser.Add_Special_Token( S : string ) ;

begin
    _Parser.Add_Special_Token( Temp_String( S ) ) ;
end ;


function TParser.Grab_Line : string ;

var Temp : PString ;

begin
    Temp := _Parser.Grab_Line ;
    Result := Temp^.Text ;
    dispose( Temp, Done ) ;
end ;


function TParser.Last_Error : longint ;

begin
    Result := _Parser.Last_Error ;
end ;


procedure TParser.Put_Token( const S : string ) ;

begin
    _Parser.Put_Token( Temp_String( S ) ) ;
end ;


function TParser.Token : string ;

var Temp : PString ;

begin
    Temp := _Parser.Token ;
    Result := Temp^.Text ;
    dispose( Temp, Done ) ;
end ;


function TParser.Token_EOL : boolean ;

begin
    Result := _Parser.Token_EOL ;
end ;



{ TFile_Parser methods... }

{ Constructors and destructors... }

constructor TFile_Parser.Create ;

begin
    inherited Create ;

    dispose( _Parser, Done ) ;
    _Parser := new( PSFile_Parser, Init ) ;
end ;


{ Property handlers... }

function TFile_Parser.Get_Source : PSFile ;

begin
    Result := PSFile_Parser( _Parser )^.Get_Source ;
end ;


procedure TFile_Parser.Set_Source( Value : PSFile ) ;

begin
    PSFile_Parser( _Parser )^.Set_Source( Value ) ;
end ;



{ TString_Parser methods... }

{ Constructors and destructors... }

constructor TString_Parser.Create ;

begin
    inherited Create ;

    dispose( _Parser, Done ) ;
    _Parser := new( PSString_Parser, Init ) ;
end ;


{ Property handlers... }

function TString_Parser.Get_Source : string ;

var Temp : PString ;

begin
    Temp := PSString_Parser( _Parser )^.Get_Source ;
    Result := Temp^.Text ;
    dispose( Temp ) ;
end ;


procedure TString_Parser.Set_Source( Value : string ) ;

var Loop : integer ;

begin
    { Since the parser strips out all CRLF, replace them with spaces so that
      tokens don't run together. }
    for Loop := 1 to length( Value ) do
    begin
        if( ( Value[ Loop ] = CR ) or ( Value[ Loop ] = LF ) ) then
        begin
            Value[ Loop ] := ' ' ;
        end ;
    end ;
    PSString_Parser( _Parser ).Set_Source( Temp_String( Value ) ) ;
end ;


{ TFiles_Parser methods...}

{ Constructors and destructors... }

constructor TFiles_Parser.Create ;

begin
    inherited Create ;

    dispose( _Parser, Done ) ;
    _Parser := new( PSFiles_Parser, Init ) ;
end ;


{ API... }

procedure TFiles_Parser.Close_File ;

begin
    PSFiles_Parser( _Parser )^.Close_File ;
end ;


procedure TFiles_Parser.Open_File( Name : string ; F : PSFile ) ;

begin
    PSFiles_Parser( _Parser )^.Open_File( Name, F ) ;
end ;
{$ENDIF}


begin
    Parse_Error := ParseErr_Success ;
end.
